#include <iostream>
#include <fstream>
#include <string>
#include <sstream>


using namespace std;


int main(int argc, char* argv[]){
	cout<<"This program is used to split paired-end sam file into two separate sam files, each corresponding to one pair-end read.\n";
	cout<<"We check the flag of the SAM file. If it's set at 0x0040, then this is the 1st read in a pair; otherwise, it's the 2nd read.\n";
	cout<<"We use this program to convert pair-end sam files into single-end sam files, which are required at step 3 of the scripture walkthrough example.\n";
	cout<<"Notice that the last character of the read name must be 1 or 2.\n";

	if(argc!=2)return -1;
	ifstream ifs(argv[1]);
	if(!ifs.is_open()){
		cerr<<"Error opening file "<<argv[1]<<endl;
		return -1;
	}
	string of1(argv[1]),of2(argv[1]), of3(argv[1]);
	of1+=".part1.sam";
	of2+=".part2.sam";
	of3+=".rename.sam";
	ofstream ofs1(of1.c_str());
	ofstream ofs2(of2.c_str());
	ofstream ofs3(of3.c_str());
	if(!ofs1.is_open() || !ofs2.is_open()){
		cerr<<"Error opening output files.\n";
		return -1;
	}

	string oneline;

	long ncount=0;
	long nunpaired=0;
	while(true){
		getline(ifs,oneline);
		//jump two lines
		if(ifs.eof())break;
		if(oneline[0]=='@')continue;//jump two header lines
		ncount++;
		if(ncount%100000==0)cout<<ncount<<" lines..."<<endl;
		stringstream ss(oneline);
		string chrname;
		ss>>chrname;
		
		//flag
		int flag;
		ss>>flag;

		//current position
		int curpos=(int)ss.tellg();

		string desname;
		ss>>desname;

		string oneline2nd=oneline.substr(curpos);//parts after flag

		if((flag & 0x01)==0){
			nunpaired++;
			continue; //ignore unpaired reads
		}
			
		int isfirst=-1;
		if (flag & 0x40)
			isfirst=1;
		else if (flag & 0x80)
			isfirst=2;

		//reset the flags
		flag &= (~ (0x01 | 0x02 | 0x08 | 0x20 | 0x40 | 0x80));
		


		//ignore the "#1" and "#2" mark
		//oneline=oneline1st+oneline2nd;
		if(isfirst==1){
			ofs1<<chrname<<desname<<"\t"<<flag<<oneline2nd<<"\n";
		}
		else if(isfirst==2){
			//ofs2<<oneline<<"\n";
			ofs2<<chrname<<desname<<"\t"<<flag<<oneline2nd<<"\n";
		}
		else{
			cerr<<"Error: incorrect read flags "<<chrname<<endl;
		}
		
		ofs3<<chrname<<desname<<"\t"<<flag<<oneline2nd<<"\n";
	}

	cout<<"Total: "<<ncount<<" reads.\n";
	cout<<"Unpaired reads: "<<nunpaired<<endl;
	ofs1.close();
	ofs2.close();
	ofs3.close();
	ifs.close();
	return 0;
}
