/*
 * mathlib.cpp
 * This file is part of isoLasso library; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2.1 of the License.
 *
 *  Created on: 2011-10-14
 *      Author: daniel
 */
#include "mathlib.h"
#include "library.h"
#include <gsl/gsl_linalg.h>
#include <algorithm>
#include <functional>
#include <math.h>
#include <typeinfo>
#include <sstream>
#include <limits>
#include <numeric>
#include <CGAL/basic.h>
#include <CGAL/QP_models.h>
#include <CGAL/QP_functions.h>

#ifdef CGAL_USE_GMP
#include <CGAL/Gmpzf.h>
typedef CGAL::Gmpzf ET;
#else
#include <CGAL/MP_Float.h>
typedef CGAL::MP_Float ET;
#endif

using namespace std;

const double PI = atan(1.)*4.;

double normpdf(double x, double mu, double sigma){
  double ff = - ((x - mu) * (x - mu));
  ff = ff / (2*sigma*sigma);
  double y = exp(ff);
  double d = sigma * sqrt((double)(2*PI));
  y = y / d;
  return y;
}

// program and solution types
typedef CGAL::Nonnegative_quadratic_program_from_iterators
<double**,                                                // for A
 double*,                                                 // for b
 CGAL::Const_oneset_iterator<CGAL::Comparison_result>, // for r
 double**,                                                // for D
 double*>                                                 // for c
Program;
typedef CGAL::Quadratic_program_solution<ET> Solution;


/*
void mldivide(vector<vector<long> > &A, vector<double> &b, vector<double> &result, bool reverseA){
  double *a_data = new double[A.size() * A[0].size()];
  double *b_data = new double[b.size()];
  for (long i = 0; i < b.size(); ++i)
    b_data[i] = b[i];
  long nRow, nCol, index = 0;
  if (reverseA == true){
    for (long j = 0; j < A[0].size(); ++j)
      for (long i = 0; i < A.size(); ++i)
        a_data[index++] = A[i][j];
    nRow = A[0].size();
    nCol = A.size();
  }
  else {
    for (long i = 0; i < A.size(); ++i)
      for (long j = 0; j < A[0].size(); ++j)
        a_data[index++] = A[i][j];
    nRow = A.size();
    nCol = A[0].size();
  }
  
  gsl_matrix_view m = gsl_matrix_view_array (a_data, nRow, nCol);

  gsl_vector_view b_value = gsl_vector_view_array (b_data, nRow);

  gsl_vector *x = gsl_vector_alloc(nCol); // size equal to n
  gsl_vector *residual = gsl_vector_alloc(nRow); // size equal to m
  gsl_vector *tau = gsl_vector_alloc (nCol); //size equal to min(m,n)
  gsl_linalg_QR_decomp (&m.matrix, tau); // 
  gsl_linalg_QR_lssolve(&m.matrix, tau, &b_value.vector, x, residual);
  result.clear();
  for (long i = 0; i < nCol; ++i)
    result.push_back(gsl_vector_get(x, i));
  gsl_vector_free (x);
  gsl_vector_free (tau);
  gsl_vector_free (residual);
  delete []a_data;
  delete []b_data;
}

void mldivide(vector<vector<double> > &A, vector<double> &b, vector<double> &result, bool reverseA){
  double *a_data = new double[A.size() * A[0].size()];
  double *b_data = new double[b.size()];
  for (long i = 0; i < b.size(); ++i)
    b_data[i] = b[i];
  long nRow, nCol, index = 0;
  if (reverseA == true){
    for (long j = 0; j < A[0].size(); ++j)
      for (long i = 0; i < A.size(); ++i)
        a_data[index++] = A[i][j];
    nRow = A[0].size();
    nCol = A.size();
  }
  else {
    for (long i = 0; i < A.size(); ++i)
      for (long j = 0; j < A[0].size(); ++j)
        a_data[index++] = A[i][j];
    nRow = A.size();
    nCol = A[0].size();
  }
  
  
  gsl_matrix_view m = gsl_matrix_view_array (a_data, nRow, nCol);

  gsl_vector_view b_value = gsl_vector_view_array (b_data, nRow);

  gsl_matrix_fprintf(stdout, &m.matrix, "%g");
  printf("---\n");
  gsl_vector_fprintf (stdout, &b_value.vector, "%g");

  gsl_vector *x = gsl_vector_alloc(nCol); // size equal to n
  gsl_vector *residual = gsl_vector_alloc(nRow); // size equal to m
  gsl_vector *tau = gsl_vector_alloc (nCol); //size equal to min(m,n)
  gsl_linalg_QR_decomp (&m.matrix, tau); // 
  printf("---\n");
  gsl_vector_fprintf (stdout, tau, "%g");
  gsl_linalg_QR_lssolve(&m.matrix, tau, &b_value.vector, x, residual);
   printf("---\n");
   gsl_vector_fprintf (stdout, residual, "%g");
   printf("---\n");
  result.clear();
  for (long i = 0; i < nCol; ++i)
    result.push_back(gsl_vector_get(x, i));
  gsl_vector_free (x);
  gsl_vector_free (tau);
  gsl_vector_free (residual);
  delete []a_data;
  delete []b_data;
}

*/

void LassoWithConstraint(vector<vector<double> > &X, vector<double> &y,double &lambda,NewInstance &ins,vector<string> &varargin, vector<double> &w, double &fval, long &exitflag){

  long minexoncount = 1, minjunccount = 1;
  exitflag = 1;
  double exonp = 1e-3, juncp = 1e-2, alpha = 5;
  bool useexoncon = false, usejunccon = true, usescale = false, boundcontrol = true;
  for (long i = 0; i < varargin.size(); ++i){
    if (varargin[i] == "--minexoncount")
      minexoncount = atol(varargin[i + 1].c_str());
    if (varargin[i] == "--minjunccount")
      minjunccount = atol(varargin[i + 1].c_str());
    if (varargin[i] == "--useexoncon"){
      useexoncon = getBoolenValue(varargin[i + 1]);
    }
    if (varargin[i] == "--usejunccon"){
      usejunccon = getBoolenValue(varargin[i + 1]);
    }
    if (varargin[i] == "--exonp")
      exonp = atof(varargin[i + 1].c_str());
    if (varargin[i] == "--juncp")
      juncp = atof(varargin[i + 1].c_str());
    if (varargin[i] == "--usescale")
      usescale = getBoolenValue(varargin[i + 1]);
    if (varargin[i] == "--alpha")
      alpha = atof(varargin[i + 1].c_str());
    if (varargin[i] == "--boundcontrol")
      boundcontrol = getBoolenValue(varargin[i + 1]);
  }

  long minexoncountinconstraint = minexoncount, minjunctioncountinconstraint = minjunccount;
  bool useexonconstraint = useexoncon, usejunctionconstraint = usejunccon;
  double pexon = exonp, pjunction = juncp;
  bool usebackslash = false;

  long m = X.size(), n = X[0].size();
  if (usebackslash){
    //mldivide(X, y, w, false);
    //exitflag = 0;
    //return;
  }
  vector<double> lb, ub,allconstb;
  vector<vector<double> >allconst;
  zeros(lb, n);
  zeros(allconst, m*(m+1)/2, n);
  zeros(allconstb, m*(m+1)/2);
  long constcount = -1;
  long genelen = accumulate(ins.exonLen.begin(), ins.exonLen.end(), 0);
  vector<double> exoncount;
  for (long i = 0; i < ins.rcount.size(); ++i)
    exoncount.push_back(((double) ins.rcountRowSum[i] + ins.rcountColSum[i]) / 2);
  
  if (useexonconstraint){
    for (long i = 0; i < m; ++i)
      if (exoncount[i] >= minexoncountinconstraint)
        if (!boundcontrol || exoncount[i]*pow(10,(double)9)/ins.totalReadCnt/ins.exonLen[i] > alpha){
          constcount++;
          allconst[constcount].assign(X[i].begin(), X[i].end());
          allconstb[constcount] = pexon;
        }
  }

  if (usejunctionconstraint){
    for (long i = 0; i < m; ++i)
      for (long j = i +1; j < m; ++j)
        if (ins.rcount[i][j] >= minjunctioncountinconstraint)
          if (!boundcontrol || exoncount[i]*pow(10, (double) 9)/ins.totalReadCnt/ins.readLen > alpha){
            constcount++;
            for (long k = 0; k < X[i].size(); ++k) allconst[constcount][k] = X[i][k] * X[j][k];
            allconstb[constcount] = pjunction;
          }
  }
/* Use this code when return to quadprog++
  QuadProgPP::Vector<double> constraintb;
  QuadProgPP::Matrix<double> constraintA;
  if (usescale){
    constraintA.resize(n, constcount + 1 + n);
    constraintb.resize(constcount + 1 + n);
    for (long i = 0; i <= constcount; ++i){
      constraintb[i] = -1 * allconstb[i];
      for (long j = 0; j < allconst[i].size(); ++j)
        constraintA[j][i] = allconst[i][j];
    }
    for (long i = constcount + 1; i <= constcount + n; ++i){
      constraintb[i] = 0;
      for (long j = 0; j < n; ++j) constraintA[j][i] = 0;
      constraintA[i - constcount - 1][i] = 1;
    }
  }
  else {
    constraintA.resize(n, 2 + constcount + n);
    constraintb.resize(2 + constcount + n);
    for (long i = 0; i < n; ++i) constraintA[i][0] = -1;
    constraintb[0] = lambda;
    for (long i = 1; i <= constcount + 1; ++i){
      constraintb[i] *= -1 * allconstb[i - 1];
      for (long j = 0; j < n; ++j)
        constraintA[j][i] = allconst[i - 1][j];
    }
    for (long i = constcount + 2; i < constcount + 2 + n; ++i){
      constraintb[i] = 0;
      for (long j = 0; j < n; ++j) constraintA[j][i] = 0;
      constraintA[i - constcount - 2][i] = 1;
    }
  }

  //vector<double> w_init;
  //mldivide(X, y, w_init, false);
  QuadProgPP::Matrix<double> H;
  QuadProgPP::Vector<double> f(X[0].size());
  matrixMul(X, X, H, true);

  if (usescale) {
    for (long j = 0; j < X[0].size(); ++j) f[j] = 0;
    for (long j = 0; j < X[0].size(); ++j)
      for (long k = 0; k < n; ++k)
        f[j] = f[j] + y[k] * X[k][j];
    for (long j = 0; j < X[0].size(); ++j) f[j] = lambda / 2 - f[j] ;
  }
  else {
    for (long j = 0; j < X[0].size(); ++j) f[j] = 0;
    for (long j = 0; j < X[0].size(); ++j){
      for (long k = 0; k < y.size(); ++k)
        f[j] = f[j] - y[k] * X[k][j];
    }
  }
  QuadProgPP::Matrix<double> CE(n, 0);
  QuadProgPP::Vector<double> ce0;

  QuadProgPP::Vector<double> x;
  x.resize(n);
  fval = QuadProgPP::solve_quadprog(H, f, CE, ce0, constraintA, constraintb, x);
  w.clear();
  for (long i = 0; i < n; ++i) {
    w.push_back(x[i]);
    if (x[i] == numeric_limits<double>::infinity()) exitflag = -1;
  }
  */
  double ** constraintA;
  double *constraintb; 
  long conssize;
  if (usescale){
    conssize = constcount + 1;
    constraintA = new double*[n];
    for (long i = 0; i < n; ++i) 
      constraintA[i] = new double[constcount + 1];
    constraintb = new double[constcount + 1];
    for (long i = 0; i <= constcount; ++i){
      constraintb[i] = -1 * allconstb[i];
      for (long j = 0; j < allconst[i].size(); ++j)
        constraintA[j][i] = -1 * allconst[i][j];
    }
  }
  else {
    conssize = constcount + 2;
    constraintA = new double*[n];
    for (long i = 0; i < n; ++i) {
      constraintA[i] = new double[constcount + 2];
      constraintA[i][0] = 1;
    }
    constraintb = new double[constcount + 2];
    constraintb[0] = lambda;
    for (long i = 1; i <= constcount + 1; ++i){
      constraintb[i] = -1 * allconstb[i - 1];
      for (long j = 0; j < allconst[i].size(); ++j)
        constraintA[j][i] = -1 * allconst[i - 1][j];
    }
  }
  double *f = new double[X[0].size()];
  double** H = new double* [X[0].size()];
  for (long i = 0; i < X[0].size(); ++i)
    H[i] = new double[i + 1];
  matrixTransposeMulSelf(X, H);


  if (usescale) {
    for (long j = 0; j < X[0].size(); ++j) f[j] = 0;
    for (long j = 0; j < X[0].size(); ++j)
      for (long k = 0; k < n; ++k)
        f[j] = f[j] + y[k] * X[k][j];
    for (long j = 0; j < X[0].size(); ++j) f[j] = lambda / 2 - f[j] ;
  }
  else {
    for (long j = 0; j < X[0].size(); ++j) f[j] = 0;
    for (long j = 0; j < X[0].size(); ++j){
      for (long k = 0; k < y.size(); ++k)
        f[j] = f[j] - y[k] * X[k][j];
    }
  }

  CGAL::Const_oneset_iterator<CGAL::Comparison_result> r(CGAL::SMALLER); // constraints are "<="
  double c0 = 0;
  w.clear();

  // now construct the quadratic program; the first two parameters are
  // the number of variables and the number of constraints (rows of A)
  try {
    Program qp (X[0].size(), conssize, constraintA, constraintb, r, H, f, c0);
    Solution s = CGAL::solve_nonnegative_quadratic_program(qp, ET());
    fval = to_double(s.objective_value());
    //printf("%lf\n", fval);
    // output solution
    CGAL::Quadratic_program_solution<ET>::Variable_value_iterator iter  = s.variable_values_begin();

    double td;
    while (iter != s.variable_values_end()){
      td = to_double(*iter);
      w.push_back(td);
      //printf("%lf\t", td);
      ++iter;
    }
    if (s.is_infeasible() == true) exitflag = -1;
  }
  catch (exception &e){
    //printf("Instance %s has exception.\n", ins.id.c_str());
    exitflag = -1;
  }
  for (long i = 0; i < X[0].size(); ++i) delete []H[i];
  delete []H;
  delete []f;
  for(int i = 0; i < n; i++) delete []constraintA[i];
  delete []constraintA;
  delete []constraintb;
  if (exitflag != 1){
    //printf("No. constraints %ld\n", constcount + 1);
    //printf("fval:  %.4lf exit flag: %ld", fval, exitflag);
    //printf("Possible erros of quadprog.\n");
  }
}
